#include "swcca.h"

#include <sys/times.h> //these two h files are for linux
#include <unistd.h>

int step;
struct tms start, stop;
int cutoff_time;

void smooth_clause_weights()
{
	int i,j,c,v;
	int new_total_weight=0;

	for (v=1; v<=num_vars; ++v) 
		dscore[v] = 0;
	
	//smooth clause dscore and update dscore of variables
	for (c = 0; c<num_clauses; ++c)
	{
		clause_weight[c] = clause_weight[c]*p_scale+scale_ave;
		
		new_total_weight+=clause_weight[c];
		
		//update dscore of variables in this clause 
		if (sat_count[c]==0) 
		{
			for(j=0; j<clause_lit_count[c]; ++j)
			{
				dscore[clause_lit[c][j].var_num] += clause_weight[c];
			}
		}
		else  if(sat_count[c]==1)
			dscore[sat_var[c]]-=clause_weight[c];
	}
	
	ave_weight=new_total_weight/num_clauses;
}

void update_clause_weights()
{
	int i,v;

	for(i=0; i < unsat_stack_fill_pointer; ++i)
		clause_weight[unsat_stack[i]]++;
	
	for(i=0; i<unsatvar_stack_fill_pointer; ++i)
	{
		v = unsatvar_stack[i];
		dscore[v] += unsat_app_count[v];
		if(dscore[v]>0 &&  conf_change[v]==1 && already_in_goodvar_stack[v] ==0)
		{
			push(v,goodvar_stack);
			already_in_goodvar_stack[v] =1;
		}
	}
	
	delta_total_weight+=unsat_stack_fill_pointer;
	if(delta_total_weight>=num_clauses)
	{
		ave_weight+=1;
		delta_total_weight -= num_clauses;
		
		//smooth weights
		if(ave_weight>threshold)
			smooth_clause_weights();
	}
}



//pick a var to be flip
int pick_var()
{
	int             i,k,c,v;
	int             best_var,best_score;
		
	/*CCD (configuration changed decreasing) mode, the level with configuation chekcing*/
	if(goodvar_stack_fill_pointer>0)
	{
		best_var = goodvar_stack[0];
		
		for(i=1; i<goodvar_stack_fill_pointer; ++i)
		{
			v = goodvar_stack[i];
			if(dscore[v] < dscore[best_var]) 
				continue;
			else if(dscore[v]>dscore[best_var])
				best_var = v;
			else if (time_stamp[v]<time_stamp[best_var])
				best_var = v;
		}
		
		return best_var;
	}
	
	/*SD (significant decreasing) mode, the level with aspiration*/
	best_var = 0;
	dscore[0] = sigscore; //GREEDY depends on ave_weight, and is changing during the search, so cannot be set in initialization

	for(i=0; i<unsatvar_stack_fill_pointer; ++i)
	{
		v = unsatvar_stack[i];
		
		if(dscore[v] < dscore[best_var]) 
			continue;
		else if(dscore[v]>dscore[best_var])
			best_var = v;
		else if (time_stamp[v]<time_stamp[best_var])
			best_var = v;
	}
		
	if(best_var != 0) 
		return best_var;
		
	//update weights
	update_clause_weights();
	
	/*random mode*/
	c = unsat_stack[rand()%unsat_stack_fill_pointer];
	
	//agest
	int oldest_var = clause_lit[c][0].var_num;
	for(k=1; k<clause_lit_count[c]; ++k)
	{
		v = clause_lit[c][k].var_num;
		if(time_stamp[v]<time_stamp[oldest_var])
			oldest_var = v;
		
	}
	
	return oldest_var;
}



void local_search(int max_flips)
{
	int flipvar;
	for (step = 0; step<max_flips; step++)
	{
	
		if(unsat_stack_fill_pointer==0) return;
		
		if(step%100000==0){// wihin cutoff_time
			times(&stop);
			
			double elap_time = (stop.tms_utime - start.tms_utime +stop.tms_stime - start.tms_stime) / sysconf(_SC_CLK_TCK);
	
			if(elap_time >= cutoff_time)return;
		}
		
		flipvar = pick_var();
		flip(flipvar);
		time_stamp[flipvar] = step;
	}
}


int main(int argc, char* argv[])
{
	int     seed, i;

	if (build_instance(argv[1])==0)
	{
		cout<<"Invalid filename: "<< argv[1]<<endl;
		return -1;
	}
     
    sscanf(argv[2],"%d",&seed);
    sscanf(argv[3],"%d",&cutoff_time);
	
	cout<<"c this is SWCCA version 2012.8.7."<<endl;
	
	srand(seed);

	times(&start);

	for (i = 1; i <= max_tries; i++) 
	{
		 init();
		 local_search(max_flips);
		 if (unsat_stack_fill_pointer==0)  break;
		 
		 times(&stop);
		 double elap_time = (stop.tms_utime - start.tms_utime +stop.tms_stime - start.tms_stime) / sysconf(_SC_CLK_TCK);
		 if(elap_time >= cutoff_time) break;
	}


	times(&stop);
	double comp_time = double(stop.tms_utime - start.tms_utime +stop.tms_stime - start.tms_stime) / sysconf(_SC_CLK_TCK);

     //print out information about the solution.
	if(unsat_stack_fill_pointer==0){
		if(verify_sol()==1)
		{
			//printf("sat %d %lf \n",step,comp_time);
			 printf("s SATISFIABLE\n");
			 print_solution();
			 printf("c tries = %d\n", i); 
			 printf("c searchSteps = %d\n", step); 
			 printf("c solveTime = %lf\n",comp_time);
		 }
	 }
	 else printf("s unkown %d %lf \n",step,comp_time);
	 
	free_memory();

    return 0;
}
